@@ -1,6 +1,6 @@
 package Devel::DumpTrace;
 
-use 5.008;
+use 5.008000;
 use Hash::SafeKeys;
 use PadWalker;
 use Scalar::Util 1.14;
@@ -8,41 +8,31 @@ use Text::Shorten;
 use Devel::DumpTrace::CachedDisplayedArray;
 use Devel::DumpTrace::CachedDisplayedHash;
 use IO::Handle;
+use File::Temp;
 use Carp;
 use strict;
 use warnings;
-# $| = 1;
 
-our $VERSION = '0.22';
+our $VERSION = '0.23';
 
 my $Time_HiRes_avail;
 my $color_avail;
-BEGIN {
-    $Time_HiRes_avail = eval 'use Time::HiRes qw(time);1' || 0;
-    $color_avail = eval
-        'use Term::ANSIColor;$Term::ANSIColor::VERSION>=3.00' || 0;
-}
-
 
 BEGIN {
+    # compile Devel::DumpTrace::Const AFTER the environment is processed
     if (defined $ENV{DUMPTRACE}) {
-	my @kv;
-	if ($ENV{DUMPTRACE} =~ /;/) {
-	    @kv = split /;/, $ENV{DUMPTRACE};
-	} else {
-	    @kv = split /,/, $ENV{DUMPTRACE};
-	}
-	foreach my $kv (@kv) {
+        my $kv_splitter = $ENV{DUMPTRACE}=~/;/ ? ';' : ',';
+	foreach my $kv (split $kv_splitter, $ENV{DUMPTRACE}) {
 	    my ($k,$v) = split /=/, $kv, 2;
 	    $ENV{"DUMPTRACE_$k"} = $v;
 	}
     }
 
-    # open a handle to the console, for debugging
-    open *Devel::DumpTrace::TTY, '>>',
-        $^O eq 'MSWin32' ? 'CON' : '/dev/tty';
+    $Time_HiRes_avail = eval 'use Time::HiRes qw(time);1' || 0;
+    $color_avail = eval
+        'use Term::ANSIColor;$Term::ANSIColor::VERSION>=3.00' || 0;
+    
 }
-# compile Devel::DumpTrace::Const AFTER the environment is processed
 use Devel::DumpTrace::Const;
 
 our $ARRAY_ELEM_SEPARATOR = ',';
@@ -62,6 +52,7 @@ our @EXCLUDE_PATTERN = ('^Devel::DumpTrace', '^Text::Shorten');
 our @INCLUDE_PATTERN = ();
 our (%DEFERRED, $PAD_MY, $PAD_OUR, $TRACE);
 our $_GLOBAL_DESTRUCTION = 0;
+our $_THREADS = 0;
 our $_INIT = 0;
 
 my (@matches, %sources);
@@ -102,18 +93,17 @@ $Devel::DumpTrace::NO_PPI
 	 	    "DUMPTRACE_FH=$ENV{DUMPTRACE_FH}. Failover to STDERR\n";
 		    $DUMPTRACE_FH = *STDERR;
 	    }
-	    $DUMPTRACE_FH->autoflush(1);
 	} else {
 	    ## no critic (BriefOpen)
 	    unless (open $DUMPTRACE_FH, '>', $ENV{DUMPTRACE_FH}) {
 		die "Can't use $ENV{DUMPTRACE_FH} as trace output file: $!\n",
 		    "Devel::DumpTrace module is quitting.\n";
 	    }
-	    $DUMPTRACE_FH->autoflush(1);
 	}
     } else {
 	$DUMPTRACE_FH = *STDERR;
     }
+    $DUMPTRACE_FH->autoflush(1);
     $DUMPTRACE_COLOR = $ENV{DUMPTRACE_COLOR} || '';
     if ($DUMPTRACE_COLOR) {
         if ($color_avail) {
@@ -144,10 +134,26 @@ sub import {
         map { '^' . $_ . '$' } split /,/, $ENV{DUMPTRACE_INCLPKG} || '';
 
     @args = grep { /^[^+-]/ } @args;
+    if (grep { $_ eq ':test' } @args) {
+
+        # :test
+        #    import some low level routines to the calling
+        #    namespace for testing.
+
+        @args = grep { $_ ne ':test' } @args;
+        no strict 'refs';
+        my $p = caller;
+        foreach my $name (qw(save_pads evaluate_and_display_line dump_scalar
+                             hash_repr array_repr handle_deferred_output
+                             evaluate save_previous_regex_matches)) {
+            *{$p . '::' . $name} = *$name;
+        }
+        *{$p . '::substitute'} = *perform_variable_substitutions;
+        *{$p . '::xsubstitute'} = *perform_extended_variable_substitutions;
+    }
     if (@args > 0) {
 	$TRACE = join ',', @args;
     }
-
     return;
 }
 
@@ -207,8 +213,6 @@ sub get_source {
 sub _exclude_pkg {
     my($file,$pkg,$line) = @_;
 
-    local *STDERR = *Devel::DumpTrace::TTY;
-
     return 0 if $INCLUDE_PKG{$pkg} || $INCLUDE_PKG{$file};
     foreach (@INCLUDE_PATTERN) {
 	if ($pkg =~ $_) {
@@ -238,6 +242,7 @@ sub _exclude_pkg {
 
 # map $TRACE variable to a display style
 sub _display_style_old {
+    return DISPLAY_TERSE if $TRACE eq 'default'; # 5.8.8 bug?
     return (DISPLAY_TERSE,
 	    DISPLAY_TERSE,
 	    DISPLAY_TERSE,
@@ -344,16 +349,20 @@ sub refresh_pads {
 }
 
 our $last_dumptrace = '';
-our @GT = ("     ",">    ",">>   ",">>>  ",">>>> ",">>>>>");
+my @dt_prefix = ("      ",">     ",">>    ",">>>   ",">>>>  ",">>>>> ",
+                 "     \t         ",">    \t         ",">>   \t         ",
+                 ">>>  \t         ",">>>> \t         ",">>>>>\t         ",
+                 "");
 
 sub dumptrace {
     my ($n, $tab, @output) = @_;
     my $dt = join ('', @output);
-    my $out = join '',($GT[$n],$tab ? "\t         " : " ", $dt);
+    my $out = $dt_prefix[$n+6*!!$tab] . $dt;
     if ($last_dumptrace && $dt eq $last_dumptrace) {
         # duplicate
         return;
     }
+
     $last_dumptrace = $dt;
     if ($DUMPTRACE_COLOR) {
         our $DUMPTRACE_RESET;
@@ -361,7 +370,19 @@ sub dumptrace {
                                split($/,$out)), "";
     }
 
-    print {$DUMPTRACE_FH} $out;
+    if ($Devel::DumpTrace::LOCKOBJ) {
+        lock(my $lock = \$Devel::DumpTrace::LOCKOBJ);
+        print {$DUMPTRACE_FH} $out;
+    } elsif ($Devel::DumpTrace::LOCKFILE) {
+        use Fcntl qw(:flock :seek);
+        open my $lock, '>>', $Devel::DumpTrace::LOCKFILE;
+        flock $lock, LOCK_EX;
+        seek $DUMPTRACE_FH, 0, SEEK_END;
+        print {$DUMPTRACE_FH} $out;
+        close $lock;
+    } else {
+        print {$DUMPTRACE_FH} $out;
+    }
 }
 
 sub evaluate_and_display_line {
@@ -382,7 +403,6 @@ sub evaluate_and_display_line {
 	my ($expr1, $op, $expr2) = ($`, $&, $');   # ');
 
 	if ($style < DISPLAY_GABBY) {
-
 	    $expr2 = perform_extended_variable_substitutions($op . $expr2, $p);
 	} else {
 	    $expr2 = perform_variable_substitutions($op . $expr2, $p);
@@ -443,9 +463,8 @@ sub evaluate_and_display_line {
 }
 
 sub separate {
-    if ($Devel::DumpTrace::SEPARATOR_USED++) {
-	print {$DUMPTRACE_FH} $SEPARATOR;
-    }
+    our $SEPARATOR_USED;
+    $SEPARATOR_USED++ && dumptrace(-1, 0, $SEPARATOR);
     return;
 }
 
@@ -698,6 +717,7 @@ sub handle_ALL_deferred_output {
 	my ($sub, $file) = split / : /, $context, 2;
 	handle_deferred_output($sub, $file);
     }
+    separate() if _display_style() > DISPLAY_TERSE;
     return;
 }
 
@@ -770,12 +790,20 @@ sub current_position_string {
 	$sub =~ s/.*:://;
 
 	if (OUTPUT_PID) {
-	    return "$$:$file:$line:[$sub]:";
+            my $p = $$;
+            if ($_THREADS) {
+                $p .= eval { "-t" . threads->tid() }; warn $@ if $@;
+            }
+	    return "$p:$file:$line:[$sub]:";
 	} else {
 	    return "$file:$line:[$sub]:";
 	}
     } elsif (OUTPUT_PID) {
-	return "$$:$file:$line:";
+        my $p = $$;
+        if ($_THREADS) {
+            $p .= eval { "-t" . threads->tid() }; warn $@ if $@;
+        }
+	return "$p:$file:$line:";
     } else {
 	return "$file:$line:";
     }
@@ -938,7 +966,7 @@ sub evaluate {
 		eval { $v = eval "\\$pkgvar" };    ## no critic (StringyEval)
 		if (!defined $v) {
 		    print {$DUMPTRACE_FH} "Devel::DumpTrace: ",
-		    "Couldn't find $sigvar/$pkgvar in any appropriate scope.\n";
+		        "Couldn't find $sigvar/$pkgvar in any appropriate scope.\n";
 		    $v = [];
 		}
 	    }
@@ -984,39 +1012,47 @@ sub save_previous_regex_matches {
     return;
 }
 
-END {
+sub __END {
     $_GLOBAL_DESTRUCTION = 1;
-    if ($$ == $pid) {
-	no warnings 'redefine';
-	handle_ALL_deferred_output();
-	separate() if _display_style() > DISPLAY_TERSE;
-	untie $TRACE;
-	$TRACE = 0;
-	*DB::DB = sub { };
-	if ($DUMPTRACE_FH ne *STDERR && $DUMPTRACE_FH ne *STDOUT) {
-	    close $DUMPTRACE_FH;
-	}
+    no warnings 'redefine';
+    *DB::DB = sub { };
+    untie $TRACE;
+    handle_ALL_deferred_output();
+}
+
+END {
+    &__END;
+}
+
+# RT#76864
+{
+    # Devel::DumpTrace is typically loaded before any other module
+    # (from the -d:DumpTrace switch). Running this  threads-specific
+    # code in a  CHECK  block gives the traced program a chance to
+    # load  threads  later.
+    no warnings 'void';
+    CHECK {
+        if ($INC{'threads.pm'}) {
+            $_THREADS = 1;
+            # idea for hijacking threads::create
+            # from  Devel::GlobalDestruction
+            my $orig_create = threads->can('create');
+            no warnings 'redefine';
+            *threads::create = sub {
+                { local $@; eval 'END { &__END }' };
+                goto $orig_create;
+            };
+            if ($INC{'threads/shared.pm'}) {
+                our $LOCKOBJ = 1;
+                threads::shared::share(\$LOCKOBJ);
+            } else {
+                our $LOCK_FH = File::Temp->new;
+                our $LOCKFILE = $LOCK_FH->filename;
+            }
+        }
     }
 }
 
-##################################################################
-
-# import Devel::DumpTrace functions to calling namespace. Used in test suite.
-sub import_all {
-    no strict 'refs';                    ## no critic (NoStrict)
-    my $p = caller;
-    *{$p . '::save_pads'} = *save_pads;
-    *{$p . '::evaluate_and_display_line'} = *evaluate_and_display_line;
-    *{$p . '::dump_scalar'} = *dump_scalar;
-    *{$p . '::hash_repr'} = *hash_repr;
-    *{$p . '::array_repr'} = *array_repr;
-    *{$p . '::handle_deferred_output'} = *handle_deferred_output;
-    *{$p . '::substitute'} = *perform_variable_substitutions;
-    *{$p . '::xsubstitute'} = *perform_extended_variable_substitutions;
-    *{$p . '::evaluate'} = *evaluate;
-    *{$p . '::save_previous_regex_matches'} = *save_previous_regex_matches;
-    return;
-}
 
 ##################################################################
 # Devel::DumpTrace::VerboseLevel: tie class for $Devel::DumpTrace::TRACE.
@@ -1028,21 +1064,18 @@ sub import_all {
 # and have the keyword 'verbose' translated into the value "5".
 #
 
-package Devel::DumpTrace::VerboseLevel;
-use Carp;
-
-sub TIESCALAR {
+sub Devel::DumpTrace::VerboseLevel::TIESCALAR {
     my ($pkg) = @_;
     my $scalar;
     return bless \$scalar, $pkg;
 }
 
-sub FETCH {
+sub Devel::DumpTrace::VerboseLevel::FETCH {
     my $self = shift;
     return ${$self};
 }
 
-sub STORE {
+sub Devel::DumpTrace::VerboseLevel::STORE {
     my ($self, $value) = @_;
 
     #Carp::cluck $self,"->STORE($value) called !\n";
@@ -1080,7 +1113,7 @@ Devel::DumpTrace - Evaluate and print out each line before it is executed.
 
 =head1 VERSION
 
-0.22
+0.23
 
 =head1 SYNOPSIS
 
@@ -1410,6 +1443,10 @@ be traced, even if it also matches one or more patterns in
 C<Devel::DumpTrace> uses the C<DUMPTRACE_FH> and C<DUMPTRACE_LEVEL>
 environment variables to configure some package variables.
 See L</"VARIABLES">.
+For Perl v5.8.8, which has a bug when the C<-d> switch is used
+like C<perl -d:Foo=xxx ...>, supplying the C<DUMPTRACE_LEVEL>
+environment variable is a workaround to this bug.
+
 The C<DUMPTRACE_NOPPI> variable can be set to force this module
 to use the basic code parser and to ignore the L<PPI|PPI>-based
 version.
@@ -1417,7 +1454,9 @@ version.
 If the C<DUMPTRACE_PID> environment variable is set to a true value,
 this module will include process ID information with the file and line
 number in all trace output. This setting can be helpful in debugging
-multi-process programs (programs that C<fork>).
+multi-process programs (programs that C<fork>). Since v0.23, the
+trace output when C<DUMPTRACE_PID> is set also includes thread ID
+information.
 
 If the C<DUMPTRACE_TIME> environment variable is set to a true value,
 this module will include program runtime information with the file
@@ -1471,7 +1510,7 @@ None known.
 
 =head1 EXPORT
 
-Nothing is or can be exported from this module.
+Nothing is exported from this module.
 
 =head1 DIAGNOSTICS
 
@@ -1642,6 +1681,10 @@ L<Tie::Trace|Tie::Trace> provides facilities to watch the values
 of specific variables, including stack trace information about
 where and how the variables values were changed.
 
+An idea for hijacking the C<threads::create> method to facilitate
+clean up of programs with threads came from the 
+L<Devel::GlobalDestruction> module.
+
 =head1 AUTHOR
 
 Marty O'Brien, E<lt>mob at cpan.orgE<gt>
