@@ -32,7 +32,7 @@ eval {use PPI;
 
 
 
-$Devel::DumpTrace::PPI::VERSION = '0.22';
+$Devel::DumpTrace::PPI::VERSION = '0.23';
 use constant ADD_IMPLICIT_ => 1;
 use constant DECORATE_FOR => 1;
 use constant DECORATE_FOREACH => 1;
@@ -42,7 +42,9 @@ use constant DECORATE_ELSIF => 1;
 use constant _DECORATE_REVERSE => 0;
 
 # built-in functions that may use $_ implicitly
-my %implicit_ = map {; $_ => 1 } qw(abs alarm chomp chop chr chroot cos
+# make %implicit_ a package, not lexical, variable on the off
+# chance that anyone wants to customize this list
+our %implicit_ = map {; $_ => 1 } qw(abs alarm chomp chop chr chroot cos
         defined eval exp glob hex int lc lcfirst length log lstat mkdir
         oct ord pos print quotemeta readlink readpipe ref require
         reverse rmdir sin split sqrt stat study uc ucfirst unlink
@@ -1469,7 +1471,7 @@ Devel::DumpTrace::PPI - PPI-based version of Devel::DumpTrace
 
 =head1 VERSION
 
-0.22
+0.23
 
 =head1 SYNOPSIS
 
